use floss::{mark_vision, Grid, Pos};
use std::io::{stdin, stdout, Write};
use termion::{
    color,
    cursor::{self, HideCursor},
    event::{Event, Key, MouseButton, MouseEvent},
    input::{MouseTerminal, TermRead},
    raw::IntoRawMode,
    screen::IntoAlternateScreen,
    style,
};

pub struct Field {
    walls: Vec<bool>,
    visible: Vec<bool>,
    width: usize,
    height: usize,
    x: usize,
    y: usize,
}

impl Grid for Field {
    fn is_wall(&self, p: Pos) -> bool {
        if let Some(i) = self.pos_index(p) {
            self.walls[i]
        } else {
            true
        }
    }

    fn mark_visible(&mut self, p: Pos) {
        if let Some(i) = self.pos_index(p) {
            self.visible[i] = true;
        }
    }
}

impl Field {
    fn new((width, height): (u16, u16)) -> Self {
        let width = width as usize / 2;
        let height = height as usize;
        let size = width * height;
        Self {
            walls: vec![false; size],
            visible: vec![false; size],
            width,
            height,
            x: width / 2,
            y: height / 2,
        }
    }

    fn pos_index(&self, (x, y): Pos) -> Option<usize> {
        if x < 0 || y < 0 {
            return None;
        }
        let x = x as usize;
        let y = y as usize;
        if x < self.width && y < self.height {
            Some(y * self.width + x)
        } else {
            None
        }
    }

    fn mouse_pos_index(&self, x: u16, y: u16) -> Option<usize> {
        self.pos_index(((x as isize - 1) / 2, y as isize - 1))
    }

    fn draw(&self) {
        let player = self.y * self.width + self.x;

        print!(
            "{}{}",
            color::Bg(color::AnsiValue::grayscale(1)),
            cursor::Goto(1, 1)
        );
        for i in 0..self.walls.len() {
            if self.visible[i] {
                print!("{}{}", color::Fg(color::LightWhite), style::Bold);
            } else {
                print!(
                    "{}{}",
                    color::Fg(color::AnsiValue::grayscale(2)),
                    style::NoBold
                );
            }

            if i == player {
                print!("@ ");
            } else if self.walls[i] {
                print!("# ");
            } else {
                print!(". ");
            }

            if (i + 1) % self.width == 0 && i < self.walls.len() - 1 {
                println!("\r");
            }
        }

        stdout().flush().unwrap();
    }

    fn render(&mut self) {
        for x in self.visible.iter_mut() {
            *x = false;
        }
        mark_vision(self, (self.x as isize, self.y as isize));
        self.draw();
    }

    fn shift(&mut self, x: isize, y: isize) {
        let new_x = self.x as isize + x;
        let new_y = self.y as isize + y;
        if !self.is_wall((new_x, new_y)) {
            self.x = new_x as usize;
            self.y = new_y as usize;
        }
    }
}

fn main() {
    let _screen = HideCursor::from(MouseTerminal::from(
        stdout()
            .into_raw_mode()
            .unwrap()
            .into_alternate_screen()
            .unwrap(),
    ));

    let mut field = Field::new(termion::terminal_size().unwrap());
    field.render();

    let mut mouse_setting = None;

    for event in stdin().events() {
        match event.unwrap() {
            Event::Key(Key::Ctrl('c')) => break,
            Event::Key(Key::Esc) => break,
            Event::Key(Key::Char('w')) => field.shift(0, -1),
            Event::Key(Key::Char('a')) => field.shift(-1, 0),
            Event::Key(Key::Char('s')) => field.shift(0, 1),
            Event::Key(Key::Char('d')) => field.shift(1, 0),
            Event::Mouse(MouseEvent::Press(MouseButton::Left, x, y)) => {
                let Some(i) = field.mouse_pos_index(x, y) else {
                    continue;
                };
                let new = !field.walls[i];
                mouse_setting = Some(new);
                field.walls[i] = new;
            }
            Event::Mouse(MouseEvent::Release(_, _)) => mouse_setting = None,
            Event::Mouse(MouseEvent::Hold(x, y)) => {
                let Some(i) = field.mouse_pos_index(x, y) else {
                    continue;
                };
                let Some(x) = mouse_setting else { continue };
                field.walls[i] = x;
            }
            _ => {}
        }
        field.render();
    }
}
